/*
 *  diskdumpfmt.c
 *
 *  Copyright (C) 2004, 2005  FUJITSU LIMITED
 *  Written by Akira Imamura (imamura.akira@jp.fujitsu.com)
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: diskdumpfmt.c,v 1.94 2005/02/25 16:39:51 akira Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include "diskdump_format.h"
#include "version.h"
#include "diskdumpcmn.h"

#define CMD	"diskdumpfmt"

static void usage(const char *cmd)
{
	fprintf(stderr, "Usage: %s [-v] [[-a] [-p] [-f]]|-c] device\n", cmd);
	fprintf(stderr, "       %s -V\n", cmd);
	fprintf(stderr, "\n");
	fprintf(stderr, "Available options:\n");
	fprintf(stderr, "\t-a:   format all blocks of the dump partition\n");
	fprintf(stderr, "\t-c:   check the status of the dump partition\n");
	fprintf(stderr, "\t-f:   format the dump partition forcibly\n");
	fprintf(stderr, "\t-p:   display progress for formatting\n");
	fprintf(stderr, "\t-v:   provide some verbose outputs\n");
	fprintf(stderr, "\t-V:   show %s version and exit\n", cmd);

	exit(255);
}

static void init_format_parameters(dump_device_t *device)
{
	int module_path;
	int num;		/* the number of parameter values gotten */

	if ((module_path = find_module_path()) < 0) {
		Err("unable to find diskdump module path.\n\
		(only full formatting is available)\n");
		return;
	}

	/* 3 values (sample rate, block order, and total blocks)
	 * are always required for quick formatting. */
	num = get_module_params(device, module_path);
	switch (num) {
	case -1:
		Err("module path specified is incorrect.\n\
		(only full formatting is available)\n");
		return;
	case 0:
		Err("diskdump module may not have been loaded.\n\
		(only full formatting is available)\n");
		return;
	case 1:
		Err("unable to get module parameter \"block order\".\n");
		exit(255);
	case 2:
		Err("unable to get module parameter \"total blocks\".\n");
		exit(255);
	case 3:
		/* got sample rate, block order, and total blocks value */
		return;	/* both full and quick formatting is available. */
	default:
		Err("returned an unexpected value '%d'.\n", num);
		exit(255);
	}
}

int main(int argc, char *argv[])
{
	dump_device_t *device;
	int opt;
	int ret;
	bool chk_flg = false, frc_flg = false, all_flg = false, prg_flg = false;
	bool verbose = false;
	off_t fmtd_blk;

	while ((opt = getopt(argc, argv, "acfpvV")) != -1) {
		switch (opt) {
		case 'a':
			all_flg = true;
			break;
		case 'c':
			chk_flg = true;
			break;
		case 'f':
			frc_flg = true;
			break;
		case 'p':
			prg_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version(CMD);
			break;
		default:
			usage(CMD);
			break;
		}
	}

	if ((chk_flg && (frc_flg || prg_flg || all_flg))) {
		usage(CMD);
	}

	if (optind + 1 != argc) {
		usage(CMD);
	}

	if ((device = alloc_device()) == NULL) {
		Err("structure allocation failed\n");
		exit(255);
	}

	if (open_device(device, argv[optind]) == FAILED) {
		Err("unable to open %s [%s]\n", device->name, device->err_msg);
		Err("note: %s\n", device->reason);
		exit(255);
	}

	if (!all_flg) {
		init_format_parameters(device);
	}

	if (chk_flg) {
		ret = check_device(device);
		show_device_info(device, ret, verbose);
		exit(ret);
	}

	if (!frc_flg) {
		if (!device->is_dump_device) {
			Err("%s: initial format is required.\n", device->name);
			exit(NOT_A_DUMP_DEVICE);
		}
	}

	if (!enough_block_size(device)) {
		Err("%s: block size should be more than %Lu\n",
			device->name, (u64)device->req->format_size);
	}

	show_format_info(device, verbose);

	if (prg_flg) {
		device->do_progress = true;
	}

	fmtd_blk = format_device(device);
	if (fmtd_blk == (off_t)-1) {
		Info("%s\n", device->reason);
		Err("formatting failed. (not suitable device)\n");
		exit(255);
	} else if (fmtd_blk != device->req->format_size) {
		Err("formatting failed. (block %Lu)\n", (u64)fmtd_blk);
		exit(255);
	} else {
		Info("formatting completed.\n");
		exit(FORMATTED);
	}
}
