/*
 *  elf.c
 *
 *  Copyright (C) 2001-2004  Red Hat, Inc
 *
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: elf.c,v 1.3 2004/10/13 19:19:24 tachino Exp $ */

#include <glib.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <endian.h>

#include "diskdumpelf.h"
#include "dumpdevice.h"

#if ELF_CLASS == ELFCLASS32

#define Elf_Nhdr	Elf32_Nhdr
#define Elf_Word	Elf32_Word
#define elf_phdr	elf32_phdr
#define Elf_Ehdr	Elf32_Ehdr
#define Elf_Phdr	Elf32_Phdr

#else

#define Elf_Nhdr	Elf64_Nhdr
#define Elf_Word	Elf64_Word
#define Elf_Ehdr	Elf64_Ehdr
#define Elf_Phdr	Elf64_Phdr

#endif

#define roundup(x, y)  ((((x)+((y)-1))/(y))*(y))

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static size_t
dump_elf_note(char *buf, Elf_Word type, char *name, char *desc, int d_len)
{
	Elf_Nhdr *note;
	size_t len;

	note = (Elf_Nhdr *)buf;
	note->n_namesz = strlen(name);
	note->n_descsz = d_len;
	note->n_type = type;
	len = sizeof(Elf_Nhdr);

	memcpy(buf + len, name, note->n_namesz);
	len = roundup(len + note->n_namesz, 4);

	memcpy(buf + len, desc, note->n_descsz);
	len = roundup(len + note->n_descsz, 4);

	return len;
}

gboolean
generate_elf_header(DumpDevice *device)
{
	char *buffer, *ptr;
	Elf_Ehdr *elf;
	Elf_Phdr *notes;
	Elf_Phdr *load;
	size_t offset, len;
	struct elf_prpsinfo prpsinfo;
	struct elf_prstatus prstatus;
	struct disk_dump_header *header = device->header;
	struct disk_dump_sub_header *sub_header = device->sub_header;
	void *task_note;
	int note_len;

	   /* should be enought for the notes + roundup + two blocks */
	buffer = g_malloc0(sizeof(Elf_Ehdr) + 2*sizeof (Elf_Phdr) +
						header->block_size * 2);

	offset = 0;
	ptr = buffer;

	/* Elf header */
	elf = (Elf_Ehdr *)ptr;
	memcpy (elf->e_ident, ELFMAG, SELFMAG);
	elf->e_ident[EI_CLASS]	= ELF_CLASS;
#if __BYTE_ORDER == __BIG_ENDIAN
	elf->e_ident[EI_DATA]	= ELFDATA2MSB;
#else
	elf->e_ident[EI_DATA]	= ELFDATA2LSB;
#endif
	elf->e_ident[EI_VERSION]= EV_CURRENT;
	elf->e_ident[EI_OSABI]	= ELFOSABI_SYSV;
	elf->e_ident[EI_ABIVERSION] = 0;
	memset(elf->e_ident+EI_PAD, 0, EI_NIDENT-EI_PAD);

	elf->e_type	= ET_CORE;
	elf->e_machine  = MACHINE_TYPE;
	elf->e_version  = EV_CURRENT;
	elf->e_entry	= 0;
	elf->e_phoff	= sizeof(Elf_Ehdr);
	elf->e_shoff	= 0;
	elf->e_flags	= 0;
	elf->e_ehsize	= sizeof(Elf_Ehdr);
	elf->e_phentsize = sizeof(Elf_Phdr);
	elf->e_phnum	= 2;
	elf->e_shentsize = 0;
	elf->e_shnum	= 0;
	elf->e_shstrndx	= 0;

	offset += sizeof(Elf_Ehdr);
	ptr += sizeof(Elf_Ehdr);

	/* PT_NOTE */

	notes = (Elf_Phdr *)ptr;
	notes->p_type	= PT_NOTE;
	notes->p_offset = 0; /* TO BE FILLED IN */
	notes->p_vaddr  = 0;
	notes->p_paddr  = 0;
	notes->p_filesz = 0; /* TO BE FILLED IN */
	notes->p_memsz  = 0;
	notes->p_flags  = 0;
	notes->p_align  = 0;

	offset += sizeof(Elf_Phdr);
	ptr += sizeof(Elf_Phdr);

	/* PT_LOAD */

	load = (Elf_Phdr *)ptr;
	load->p_type   = PT_LOAD;
	load->p_offset = 0; /* TO BE FILLED IN */
	load->p_vaddr  = device->page_offset;
	load->p_paddr  = 0;
	load->p_filesz = header->block_size * header->max_mapnr;
	load->p_memsz  = load->p_filesz;
	load->p_flags  = PF_R|PF_W|PF_X;
	load->p_align  = header->block_size;

	offset += sizeof(Elf_Phdr);
	ptr += sizeof(Elf_Phdr);

	notes->p_offset = offset;

	/* NT_PRSTATUS note */
	memset(&prstatus, 0, sizeof(prstatus));
	memcpy(&prstatus.pr_reg, sub_header->elf_regs, sizeof(prstatus.pr_reg));
	len = dump_elf_note(ptr, NT_PRSTATUS, "CORE",
				(char *)&prstatus, sizeof(prstatus));
	offset += len;
	ptr += len;
	notes->p_filesz += len;


	/* NT_PRPSINFO note */
	memset(&prpsinfo, 0, sizeof(struct elf_prpsinfo));
	prpsinfo.pr_state	= 0;
	prpsinfo.pr_sname	= 'R';
	prpsinfo.pr_zomb	= 0;
	strcpy(prpsinfo.pr_fname, "vmlinux");

	len = dump_elf_note(ptr, NT_PRPSINFO, "CORE",
				(char *)&prpsinfo, sizeof(prpsinfo));

	offset += len;
	ptr += len;
	notes->p_filesz += len;

	/* NT_TASKSTRUCT note */
	task_note = get_task_struct_note(device, &note_len);
	len = dump_elf_note(ptr, NT_TASKSTRUCT, "CORE", task_note, note_len);
	g_free(task_note);
	offset += len;
	ptr += len;
	notes->p_filesz += len;

	/* Page align offset */
	offset = roundup(offset, header->block_size);

	load->p_offset = offset;
	device->data_offset = offset;

	while (offset > 0) {
		len = write(device->file_fd,
			buffer + (device->data_offset - offset), offset);
		if (len < 0) {
			g_free(buffer);
			return TRUE;
		}

		offset -= len;
	}

	g_free(buffer);
	return FALSE;
}
