/*
 *  savecore.c
 *
 *  Copyright (C) 2004, 2005  FUJITSU LIMITED
 *  Copyright (C) 2001-2004  Red Hat, Inc
 *
 *  Written by Nobuhiro Tachino (tachino@jp.fujitsu.com) based on
 *  netdump-server code.
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: savecore.c,v 1.35 2005/03/01 16:14:21 akira Exp $ */

#include <glib.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#include "diskdumpelf.h"
#include "dumpdevice.h"
#include "version.h"
#include "diskdumpcmn.h"

#define CMD	"savecore"

int prog_flg = 0;

gboolean generate_elf_header(DumpDevice *);

/*
 * We save the disk dump under /var/crash/127.0.0.1-<date>/..
 */
#define DISKDUMP_DIR_PREFIX "/var/crash"
#define DISKDUMP_IP	0x7f000001

static DumpDevice *create_new_device(const char *device_name)
{
	DumpDevice *device;

	device = g_new0(DumpDevice, 1);

	device->device_name = g_strdup(device_name);

	device->file_fd = -1;
	device->device_fd = -1;

	/* Lame default. should be read from kernel */
	device->page_offset = PAGE_OFFSET;

	return device;
}

static int open_device_file(DumpDevice *device)
{
	int fd;

	fd = open(device->device_name, O_RDONLY);
	if (fd < 0) {
		Syserr("unable to open device file %s", device->device_name);
		return FAILED;
	}
	device->device_fd = fd;
	return SUCCEEDED;
}

static int execute_script(const char *script, const char *dir)
{
	struct stat sbuf;
	char *filename;
	char *cmdline;
	int res = -1;

	filename = g_strconcat(DISKDUMP_DIR_PREFIX "/scripts/",
			  script,
			  NULL);

	if (stat(filename, &sbuf) == 0) {
		cmdline = g_strdup_printf("%s '%s'",
					 filename, (dir)?dir:"");
		res = system(cmdline);
		g_free(cmdline);
	}

	g_free(filename);
	return res;
}

static int check_fs(DumpDevice *device)
{
	struct statfs buf;
	struct disk_dump_header *header = device->header;
	off_t memdump_size;
	int ret;
	gboolean tried_once = FALSE;

redo:
	if (fstatfs(device->file_fd, &buf) < 0) {
		Err("unable to statfs on the crash dump file\n");
		return FAILED;
	}

	memdump_size = (off_t)header->written_blocks + 3; /* elf header */
	memdump_size *= header->block_size;
	memdump_size = (memdump_size + (buf.f_bsize - 1)) / buf.f_bsize;

	if (memdump_size <= buf.f_bavail)
		return SUCCEEDED;
	if (!tried_once++) {
		ret = execute_script("diskdump-nospace", device->dump_dir);
		if (ret > 0)
			goto redo;
	}

	Err("no space for dump image\n");
	return FAILED;
}

static const char *get_dir(DumpDevice *device)
{
	char *path;
	time_t timeofdeath;
	int res;
	guint32 ip = DISKDUMP_IP;
	struct tm *tm;
	struct disk_dump_header *header = device->header;
	char date[128];

	timeofdeath = (time_t)header->timestamp.tv_sec;

	tm = localtime(&timeofdeath);
	strftime(date, sizeof(date), "%Y-%m-%d-%H:%M", tm);
	path = g_strdup_printf("%s/%d.%d.%d.%d-%s",
			       DISKDUMP_DIR_PREFIX,
			       (ip >> 24) & 0xff,
			       (ip >> 16) & 0xff,
			       (ip >> 8) & 0xff,
			       (ip) & 0xff,
			       date);

	res = mkdir(path, S_IRWXU);
	if (res) {
		Err("error making crashdump directory %s\n", path);
		g_free(path);
		return NULL;
	}

	return path;
}

static int open_dump_file(DumpDevice *device)
{
	const char *dir;
	char *file_name;
	int fd;

	dir = get_dir(device);
	if (dir == NULL)
		return FAILED;

	device->dump_dir = dir;

	file_name = g_strconcat(device->dump_dir, "/vmcore-incomplete", NULL);
	fd = open(file_name, O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR);
	if (fd < 0) {
		Syserr("unable to open crash dump file %s", file_name);
		g_free(file_name);
		return FAILED;
	}

	device->file_fd		= fd;
	device->file_name	= file_name;

	return SUCCEEDED;
}

static int read_device_header(DumpDevice *device)
{
	struct disk_dump_header *header;
	struct disk_dump_sub_header *sub_header;
	int ret;
	int bitmap_len;
	int block_size;
	int sub_header_len;

	block_size = (int)sysconf(_SC_PAGESIZE);

	header = g_malloc(block_size);

	lseek(device->device_fd, block_size, SEEK_SET);

	ret = read(device->device_fd, header, block_size);
	if (ret < 0 || ret != block_size) {
		Err("cannot read dump header\n");
		return FAILED;
	}

	/* validate dump header */
	if (memcmp(header->signature, DISK_DUMP_SIGNATURE,
				sizeof(header->signature))) {
		Err("device does not have panic dump header\n");
		return FAILED;
	}
	if (header->block_size != block_size) {
		Err("block size in the dump header does not match with system page size\n");
		return FAILED;
	}
	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size ||
	    header->nr_cpus <= 0) {
		Err("invalid nr_cpus value: %d\n", header->nr_cpus);
		return FAILED;
	}

	/* read sub header */
	lseek(device->device_fd, block_size * 2, SEEK_SET);
	sub_header_len = sub_header_length(header);
	sub_header = g_malloc(sub_header_len);
	ret = read(device->device_fd, sub_header, sub_header_len);
	if (ret < 0 || ret != sub_header_len) {
		Err("cannot read dump sub header\n");
		return FAILED;
	}
	device->sub_header = sub_header;

	/* read memory bitmap */
	bitmap_len = block_size * header->bitmap_blocks;

	lseek(device->device_fd, block_size * (2 + header->sub_hdr_size), SEEK_SET);
	device->bitmap = g_malloc(bitmap_len);
	ret = read(device->device_fd, device->bitmap, bitmap_len);
	if (ret < 0 || ret != bitmap_len) {
		Err("cannot read memory bitmap\n");
		return FAILED;
	}


	/*
	 * page 0: device header
	 * page 1: dump header
	 * page 2..: dump sub header
	 * page 2 + size of subheader: memory page 0
	 */
	device->device_page_start = (2 + header->sub_hdr_size + header->bitmap_blocks) * header->block_size;

	device->header = header;

	return SUCCEEDED;
}

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static int write_dump_header(DumpDevice *device)
{
	int err;

	err = generate_elf_header(device);
	if (err)
		return FAILED;
	else
		return SUCCEEDED;
}

static inline gboolean check_zero_page(DumpDevice *device, char *data)
{
	int i;

	for (i = 0; i < device->header->block_size; i++)
		if (data[i])
			return FALSE;
	return TRUE;
}

static inline int page_is_ram(int nr, DumpDevice *device)
{
	return device->bitmap[nr>>3] & (1 << (nr & 7));
}

static void _print_progress(const char *device_name, int blk_nr, int blk_max)
{
	double progress;

	progress = (double)blk_nr * 100.0
		 / (double)blk_max;

	if (prog_flg)
		fprintf(stderr, "%s: %5.1f%%\r", device_name, progress);
}

static void print_progress(const char *device_name, int blk_nr, int blk_max)
{
	time_t tm;
	static time_t last_time = 0;

	tm = time(NULL);

	if (tm - last_time < 1)
		return;

	last_time = tm;

	_print_progress(device_name, blk_nr, blk_max);
}

static void remain_progress(const char *device_name, int blk_nr, int blk_max)
{
	_print_progress(device_name, blk_nr, blk_max);
	if (prog_flg)
		fprintf(stderr, "\n");
}


static int write_pages(DumpDevice *device)
{
	int len;
	int i;
	char *data;
	int ret = 0;
	struct disk_dump_header *header = device->header;
	int block_size = header->block_size;

	data = g_malloc(block_size);

	lseek(device->device_fd, (off_t)device->device_page_start, SEEK_SET);
	lseek(device->file_fd, (off_t)device->data_offset, SEEK_SET);

	for (i = 0; i < header->max_mapnr; i++) {
		print_progress(device->device_name, i, header->max_mapnr);
		if (page_is_ram(i, device)) {
			len = read(device->device_fd, data, block_size);
			if (len != block_size) {
				remain_progress(device->device_name, i,
						header->max_mapnr);
				Err("device read failed\n");
				ret = -1;
				goto out;
			}
			if (check_zero_page(device, data)) {
				lseek(device->file_fd, block_size, SEEK_CUR);
				continue;
			}
			len = write(device->file_fd, data, block_size);
			if (len != block_size) {
				remain_progress(device->device_name, i,
						header->max_mapnr);
				Syserr("error writing memory dump");
				ret = -1;
				goto out;
			}
		} else {
			lseek(device->file_fd, block_size, SEEK_CUR);
		}
	}
	remain_progress(device->device_name, i, header->max_mapnr);

	/*
	 * Enforce to set the size of vmcore because last block may not be
	 * actually written.
	 */
	ret = ftruncate(device->file_fd, (off_t)device->data_offset +
					 (off_t)header->max_mapnr * block_size);
	if (ret < 0) {
		perror("ftruncate");
		goto out;
	}
out:
	g_free(data);
	return ret;
}

static int rename_vmcore(DumpDevice *device)
{
	char *old_file;
	char *new_file;

	old_file = g_strconcat(device->dump_dir, "/vmcore-incomplete", NULL);
	new_file = g_strconcat(device->dump_dir, "/vmcore", NULL);

	if (link(old_file, new_file) == 0)
		unlink(old_file);

	g_free(old_file);
	g_free(new_file);

	return SUCCEEDED;
}

static void usage(const char *cmd) {
	fprintf(stderr, "Usage: %s [-p] device\n", cmd);
	fprintf(stderr, "       %s -V\n", cmd);
	fprintf(stderr, "\n");
	fprintf(stderr, "Available options:\n");
	fprintf(stderr, "\t-p:   display progress for writing memory dump\n");
	fprintf(stderr, "\t-V:   show %s version and exit\n", cmd);

	exit(1);
}

int main(int argc, char *argv[])
{
	DumpDevice *device;
	struct disk_dump_header *header;
	int opt;
	int ret;
	unsigned int total_blocks;

	while ((opt = getopt(argc, argv, "pV")) != -1) {
		switch (opt) {
		case 'p':
			prog_flg = 1;
			break;
		case 'V':
			version(CMD);
			break;
		default:
			usage(CMD);
			break;
		}
	}

	if (optind + 1 != argc)
		usage(CMD);

	device = create_new_device(argv[optind]);
	if (device == NULL)
		exit(1);
	if ((ret = open_device_file(device)) < 0)
		exit(1);
	if ((ret = read_device_header(device)) < 0)
		exit(1);
	if ((ret = open_dump_file(device)) < 0)
		exit(1);
	if ((ret = check_fs(device)) < 0)
		exit(1);
	if ((ret = write_dump_header(device)) < 0)
		exit(1);
	if ((ret = write_pages(device)) < 0)
		exit(1);

	header = device->header;

	total_blocks =    1 /* partition header */
			+ 1 /* common header */
		        + header->sub_hdr_size
			+ header->bitmap_blocks
			+ header->total_ram_blocks;

	if (total_blocks != header->written_blocks) {
		Err("dump partition has only %u pages. %u expected\n",
			header->written_blocks, header->total_ram_blocks);
		exit(1);
	} else {
		if ((ret = rename_vmcore(device)) != 0)
			exit(1);
	}

	exit(0);
}
